"""Init file for the networks module."""

from experiment_lab.common.networks.network import (
    create_network,
    create_aggregated_network,
    create_multi_network,
)
from experiment_lab.common.networks.mlp import create_mlp_network
